module Optimize where

import Prelude hiding (catch)
import Concurrent
import Exception
import Maybe (isJust)
import List
import IO
import SML

type Optimizer a = a -> Document -> Maybe (a, Document)
type Path = [Int]
type OptimizerTable = [ (String, Document -> Maybe Document) ]

optimizerList :: OptimizerTable
optimizerList = [ ("removeIneffective", removeIneffective),
                  ("mergeTexts", allChildrenMaybe mergeTexts),
                  ("mergeWSNodes", allChildrenMaybe mergeWSNodes),
                  ("mergeNestedHiddenEffect", allChildrenMaybe mergeNestedHiddenEffect),
                  ("removeMatchingEm", removeMatchingEm ),
                  ("mergeSiblings", allChildrenMaybe mergeSiblings),
                  ("mergeRightWS", allChildrenMaybe mergeRightWS),
                  ("mergeLeftWS", allChildrenMaybe mergeLeftWS),
                  ("mergeHiddenEffect", allChildrenMaybe mergeHiddenEffect),
                  ("optimizeDocSeq", allChildrenMaybe $ optimizeDocSeq 3 2) ]

runOptimizerList :: OptimizerTable -> MVar Document -> Document -> IO Bool
runOptimizerList [] result doc = return False
runOptimizerList ((name, x):xs) result doc = do
   hPutStrLn stderr ("Running: " ++ name)
   res <- (return $! (x doc))
             `Exception.catch` (\e -> case e of
                                      AsyncException ThreadKilled -> throw e
                                      _ -> do hPutStrLn stderr ("Optimizer exception: " ++ show e)
                                              return Nothing)
   case res of
      Nothing   -> do hPutStrLn stderr ("(no result)")
                      runOptimizerList xs result doc
      Just doc' -> do updateResult name result doc'
                      runOptimizerList xs result doc'
                      return True

loopRunOptimizerList :: Int -> OptimizerTable -> MVar Document -> Document -> IO ()
loopRunOptimizerList cost opts result doc 
    = do res <- runOptimizerList opts result doc
         case res of
          False -> do let depth = cost
                      let lookAhead = max 3 (30 `div` (depth + 4))
                      if depth > docDepth doc then 
                          return ()
                       else do
                          res <- runOptimizerList [ ("optimizeDocSeq " ++ show lookAhead ++ " " ++ show depth,
                                                    allChildrenMaybe $ optimizeDocSeq lookAhead depth) ]
                                                  result
                                                  doc
                          case res of 
                           False -> loopRunOptimizerList (cost + 1) opts result doc
                           True  -> do doc' <- readMVar result
                                       loopRunOptimizerList (min 1 (cost - 1)) opts result doc'
          True  -> do doc' <- readMVar result
                      loopRunOptimizerList cost opts result doc'

optimizeDocument :: MVar Document -> IO ()
optimizeDocument result = do
    doc <- readMVar result
    loopRunOptimizerList 1 optimizerList result doc
    return ()

--      runOptimizer topDownRemoveIneffective doc doc
--      where
--      runOptimizer f init doc = 
--          case f init doc of
--              Nothing        -> return ()
--              Just (s, doc') -> do updateResult result doc'
--                                   runOptimizer f s doc'

mergeTexts :: Document -> Maybe Document
mergeTexts = merge2 (\_ p1 p2 -> isText p1 && isText p2) mergeText

mergeWSNodes :: Document -> Maybe Document
mergeWSNodes doc@(Tag ctx tag [child])
   | isAllWS child && isSpaceContextEq ctx (context child) 
       = Just (child { context = ctx })
   | otherwise = Nothing
mergeWSNodes ps = Nothing

mergeNestedHiddenEffect :: Document -> Maybe Document
mergeNestedHiddenEffect (Tag ctx t [child])
   | not (isText child) && hideEffectOf (tag child) t
       = Just (child { context = ctx })
   | otherwise = Nothing
mergeNestedHiddenEffect _ = Nothing

isAllWS :: Document -> Bool
isAllWS (Text ctx txt) = and (map isWhiteSpace txt)
isAllWS _              = False

endsInWS :: Document -> Bool
endsInWS (Text ctx txt) = isWhiteSpace (last txt)
endsInWS _              = False

startsWithWS :: Document -> Bool
startsWithWS (Text ctx txt) = isWhiteSpace (head txt)
startsWithWS _              = False

mergeSiblings :: Document -> Maybe Document
mergeSiblings = 
    merge2 (\ _ p1 p2 -> 
            not (isText p1) && not (isText p2) && (tag p1 == tag p2))
           mergeTags

merge2 :: (Document -> Document -> Document -> Bool)
          -> (Document -> Document -> Document)
          -> Document
          -> Maybe Document
merge2 pred merger v@(Tag ctx t (p1:p2:ps))
    | pred v p1 p2
        = case merge2 pred merger (Tag ctx t (merger p1 p2 : ps)) of
          Nothing -> Just (Tag ctx t (merger p1 p2 : ps))
          def     -> def
    | otherwise 
        = case merge2 pred merger (Tag ctx t (p2:ps)) of
          Nothing -> Nothing
          Just n  -> Just (updateChildren n (p1 : children n))
merge2 _ _ _ = Nothing

merge3 :: (Document -> Document -> Document -> Document -> Bool)
          -> (Document -> Document -> Document -> Document)
          -> Document
          -> Maybe Document
merge3 pred merger v@(Tag ctx t (p1:p2:p3:ps))
    | pred v p1 p2 p3
        = case merge3 pred merger (Tag ctx t (merger p1 p2 p3 : ps)) of
          Nothing -> Just (Tag ctx t (merger p1 p2 p3 : ps))
          def     -> def
    | otherwise 
        = case merge3 pred merger (Tag ctx t (p2:p3:ps)) of
          Nothing -> Nothing
          Just n  -> Just (updateChildren n (p1 : children n))
merge3 _ _ _ = Nothing

mergeHiddenEffect :: Document -> Maybe Document
mergeHiddenEffect = 
    merge3 (\ _ p1 p2 p3 -> 
            not (isText p1) && not (isText p2) && not (isText p3)
            && tag p1 == tag p3 && hideEffectOf (tag p2) (tag p1))
           (\ p1 p2 p3 -> 
            Tag (context p1) (tag p1) (children p1 ++ [p2 { context = childContext p1 }] ++ children p3))

mergeTags :: Document -> Document -> Document
mergeTags (Tag ctx t cs1) (Tag _ _ cs2) = (Tag ctx t (cs1 ++ cs2))
mergeTags _ _ = error "mergeTags not called on tags" 

mergeLeftWS :: Document -> Maybe Document
mergeLeftWS =
    merge2 (\_ p1 p2 -> not (isText p1) && isAllWS p2 && isSpaceContextEq (context p1) (childContext p1))
           (\p1 p2 -> updateChildren p1 (children p1 ++ [updateTag p2 (tag p1)]))

mergeRightWS :: Document -> Maybe Document
mergeRightWS =
    merge2 (\_ p1 p2 -> not (isText p2) && isAllWS p1 && isSpaceContextEq (context p2) (childContext p2))
           (\p1 p2 -> updateChildren p2 (updateTag p1 (tag p2) : children p2))

allChildren :: ([Document] -> [Document]) -> Document -> Document
allChildren optimizer doc
   | isText doc = doc
   | otherwise  = let cs = optimizer (map (allChildren optimizer) (children doc))
                  in updateChildren doc cs
                      

allChildrenMaybe :: (Document -> Maybe Document) -> Document -> Maybe Document
allChildrenMaybe optimizer doc
   | isText doc = Nothing
   | otherwise
       = let origChildren = children doc
             updatedChildren = map (allChildrenMaybe optimizer) 
                                   origChildren
             change = any isJust updatedChildren
             doc' = if change then 
                        updateChildren doc 
                                       (updateContexts (childContext doc)
                                                       (updateJust origChildren 
                                                                   updatedChildren))
                    else
                       doc
        in  case optimizer doc' of
            Nothing -> if change then Just doc' else Nothing
            def     -> def

updateJust :: [a] -> [Maybe a] -> [a]
updateJust orig updated =
    [ maybe origElement id updatedElement
    | (origElement, updatedElement) <- zip orig updated ]

updateJustMany :: [a] -> [Maybe [a]] -> [a]
updateJustMany orig updated 
    = concat (updateJust (map (:[]) orig) updated)

updateResult :: String -> MVar Document -> Document -> IO ()
updateResult name var new = 
    block (do old <- takeMVar var
              compareDocs name old new
              putMVar var new)

compareDocs :: String -> Document -> Document -> IO ()
compareDocs name old new = do
    let lold = docSize old
    let lnew = docSize new
    hPutStrLn stderr (name ++ ": old length: " ++ show lold ++ ", new length: " ++ show lnew)
    hPutStrLn stderr ("old: " ++ (take 100 (show old)))
    hPutStrLn stderr ("new: " ++ (take 100 (show new)))

removeIneffective :: Document -> Maybe Document
removeIneffective parent 
    = case parent of
          Tag c t children ->
              let updated = [ removeIneffectiveTag (applyTag c t) d | d <- children ]
                  changed = any isJust updated
              in  if changed then
                      Just (Tag c t (updateContexts (applyTag c t) (updateJustMany children updated)))
                  else
                      Nothing
          Text _ _ -> Nothing

removeIneffectiveTag :: Context -> Document -> Maybe [Document]
removeIneffectiveTag c (Tag c' t children)
    | isEffective c t = Nothing
    | otherwise       = Just children
removeIneffectiveTag _ _ = Nothing

removeMatchingEm :: Document -> Maybe Document
removeMatchingEm parent
    = case parent of
          Tag c t children -> 
              let updated = [ removeMatchingEmIfEm d | d <- children ]
                  changed = any isJust updated
              in  if changed then
                      Just (Tag c t (updateContexts (applyTag c t) (updateJustMany children updated)))
                  else
                      Nothing
          Text _ _ -> Nothing

removeMatchingEmIfEm :: Document -> Maybe [Document]
removeMatchingEmIfEm current
    = case current of
          Tag ctx EM children -> 
              case removeMatchingEm' ctx children of
              Nothing -> continue 
              def -> def
          Tag _ _ _ -> continue
          _ -> Nothing
    where continue = case removeMatchingEm current of
                     Nothing -> Nothing
                     Just d -> Just [d]

removeMatchingEm' :: Context -> [Document] -> Maybe [Document]
removeMatchingEm' parentContext [] = Nothing
removeMatchingEm' parentContext [child]
   = case branchPredicate matchingEm child [] of
     [] -> Nothing
     (match:visited) -> 
         case match of 
         Tag _ EM _ -> Just (rebuildTopDown parentContext
                                            (reverse visited)
                                            (children match))
         Tag _ S  _ -> Just (rebuildTopDown parentContext
                                            (reverse visited)
                                            [match])
         _          -> error "should be EM or S"

removeMatchingEm' _ ps = Nothing


rebuildTopDown :: Context -> [Document] -> [Document] -> [Document]
rebuildTopDown ctx (p:ps) children
    = [updateContext ctx 
                     (updateChildren p 
                                     (rebuildTopDown (applyTag ctx (tag p)) 
                                                     ps children))]
rebuildTopDown ctx [] children
    = map (updateContext ctx) children

rebuildPath :: [Document] -> [Document] -> [Document]
rebuildPath children [] = children
rebuildPath children (p:ps)
    = rebuildPath [updateChildren p children] ps

data SearchStatus = Found | Continue | Stop

matchingEm :: Document -> SearchStatus
matchingEm (Tag _ EM children) = Found
matchingEm (Tag _ S  children) = Found
matchingEm (Tag _ PL children) = Stop
matchingEm _                   = Continue

topDownRemoveIneffective :: Optimizer Document
topDownRemoveIneffective current root 
    = case current of
          Tag _ _ [] -> Nothing
          Tag _ _ (x:xs) -> Just (x, root)
          Text _ s   -> Nothing

branchPredicate :: (Document -> SearchStatus) -> Document -> [Document] -> [Document]
branchPredicate p branch visited
    = case p branch of
      Found    -> branch:visited
      Stop     -> []
      Continue ->
          case branch of
          Tag _ _ [child] -> branchPredicate p child (branch:visited)
          _        -> []


------------------------------------------------------

--  replaceDocSequence :: Context -> [(Document, Context)] -> [Document]
--  replaceDocSequence context [] = []
--  replaceDocSequence context docs
--       = [ makeDoc trans doc | (tags, doc) <- zip (makeTransitions docs) docs ]

expandChildren :: Int -> Document -> [(Document, Context)]
expandChildren 0 v@(Text ctx s) = [(v, ctx)]
expandChildren 0 v@(Tag ctx t children) 
    = [ (child, applyTag ctx t) | child <- children ]
expandChildren n v@(Text ctx s) = [(v, ctx)]
expandChildren n v@(Tag ctx t children)
    = concat [ expandChildren (n - 1) child | child <- children ]

makeContext :: Document -> (Document, Context)
makeContext v@(Text ctx s) = (v, ctx)
makeContext v@(Tag ctx t _) = (v, applyTag ctx t)


makeContexts :: [Document] -> [(Document, Context)]
makeContexts docs = map makeContext docs

optimizeDocSeq :: Int -> Int -> Document -> Maybe Document
optimizeDocSeq lookahead depth (Text _ _) = Nothing
optimizeDocSeq lookahead depth (Tag ctx t children)
    = let docs = [ doc | child <- children, doc <- expandChildren depth child ]
          new  = map fst (optimizeDocSeqLimit lookahead (applyTag ctx t) docs)
          s1   = sum (map docSize children)
          s2   = sum (map docSize new)
      in  case s1 <= s2 of
          True -> Nothing
          False -> Just (Tag ctx t new)

optimizeDocSeqLimit :: Int -> Context -> [(Document, Context)] -> [(Document, Context)]
optimizeDocSeqLimit n ctx [] = []
optimizeDocSeqLimit n ctx docs
        = let init = take n docs
              rest = drop n docs
          in case optimizeDocSequence ctx init of
             [] -> head docs : optimizeDocSeqLimit n ctx (tail docs)
             opt -> (head opt) : optimizeDocSeqLimit n ctx (tail (opt ++ rest))

optimizeDocSequence :: Context -> [(Document, Context)] -> [(Document, Context)]
optimizeDocSequence ctx docs 
    = case (replaceDocSequence ctx [] docs) of
      [] -> error "no path found!"
      paths -> fst (buildDocs ctx (minimumBy minTagTransition paths))

replaceDocSequence :: Context -> [(Tags, Context)] -> [(Document, Context)] -> [[TagOp]]
replaceDocSequence from open [] = [[]]
replaceDocSequence from open ((doc, to):rest)
    = [ trans ++ [Doc doc] ++ rests 
      | trans <- findMultiTagTransition open from to,
        rests <- replaceDocSequence to (applyTrans from open trans) rest ]

buildDocs :: Context -> [TagOp] -> ([(Document, Context)], [TagOp])
buildDocs ctx []           = ([], [])
buildDocs ctx xs@(Pop _:_) = ([], xs)
buildDocs ctx xs = case buildDoc ctx xs of
                   (doc, rest) -> 
                       case buildDocs ctx rest of
                       (docs, rest') -> (doc ++ docs, rest')

buildDoc :: Context -> [TagOp] -> ([(Document, Context)], [TagOp])
buildDoc context [] = ([], [])
buildDoc context xs@(Pop _:_) = ([], xs)
buildDoc context (Push t:rest) = buildTag context t rest
buildDoc context (Doc d:rest)  = ([(d, context)], rest)

buildTag :: Context -> Tags -> [TagOp] -> ([(Document, Context)], [TagOp])
buildTag context tag input
    = case applyTag context tag of
      context' ->
          case buildDocs context' input of
          (docs, []) -> ([(Tag context tag (map fst docs), context)], [])
          (docs, rest) -> ([(Tag context tag (map fst docs), context)], tail rest)

