module Parser (parseSML, mainParser) where

import Char (ord)
import List (find)
import Maybe (isJust)
import Parsec
import SML
--  import UnitTest

parseSML :: Parser Document
parseSML = do docs <- parseDocs rootContext
              eof
              return (Tag rootContext Root docs)

parseTag :: Context -> Tags -> Parser [Document]
parseTag context tag = do
    string (show tag)
    char '>'
    case applyTag context tag of
        context' -> do
            docs <- parseDocs context'
            string "</"
            string (show tag)
            char '>'
            return (makeTag context tag docs)

parseAnyTag :: Context -> Parser [Document]
parseAnyTag context = do
    try (do char '<'; notFollowedBy (char '/'))
    choice (map (parseTag context) allTags)

parseDocs :: Context -> Parser [Document]
parseDocs context = do d <- many (parseDoc context)
                       return (concat d)

parseDoc :: Context -> Parser [Document]
parseDoc context = ((parseAnyTag context <|> parseText context)
                    <?> "valid tag or text")

parseText :: Context -> Parser [Document]
parseText context 
    = do text <- many1 (satisfy isTextChar)
         return [makeText context text]


mainParser :: String -> Document
mainParser input = let (docs, []) = pDocs rootContext input
                   in  Tag rootContext Root docs

pDocs :: Context -> String -> ([Document], String)
pDocs context [] = ([], "")
pDocs context xs@('<':'/':_) = ([], xs)
pDocs context xs = let (doc, rest) = pDoc context xs
                       (docs, rest') = pDocs context rest
                   in (doc ++ docs, rest')

pDoc :: Context -> String -> ([Document], String)
pDoc context [] = error "expected document"
pDoc context xs@('<':'/':_) = ([], xs)
pDoc context ('<':rest)     = pAnyTag context rest
pDoc context xs@(x:_)       = pText context xs

matchS :: String -> String -> Maybe String
matchS text input = let text' = take (length text) input
                        rest = drop (length text) input
                    in if text == text' then Just rest else Nothing

pAnyTag :: Context -> String -> ([Document], String)
pAnyTag context input = case find isJust (pTags context input) of
                        Nothing -> error "expected valid tag"
                        Just (Just res) -> res

pTags :: Context -> String -> [Maybe ([Document], String)]
pTags context input = 
    map (pTag context input) allTags

pTag :: Context -> String -> Tags -> Maybe ([Document], String)
pTag context input tag 
    = case matchS startTag input of
      Nothing   -> Nothing
      Just rest -> let context' = applyTag context tag
                       (docs, rest') = pDocs context' rest
                   in case matchS endTag rest' of
                      Nothing     -> error ("expected: " ++ endTag)
                      Just rest'' -> Just (makeTag context tag docs, rest'')
    where startTag = show tag ++ ">"
          endTag = "</" ++ show tag ++ ">"

pText :: Context -> String -> ([Document], String)
pText context xs = context `seq` xs `seq`
    case takeWhile isTextChar xs of
    text -> case dropWhile isTextChar xs of
            rest -> ([makeText context text], rest)

isTextChar :: Char -> Bool
isTextChar '\x20' = True
isTextChar '\x0d' = True
isTextChar '\x0a' = True
isTextChar '\x09' = True
isTextChar '<'    = False
isTextChar '>'    = False
isTextChar c      = ord c >= 0x21 && ord c <= 0x7e

-- Tests

--  testParser :: TestCase ()
--  testParser = return () 
-- testSuite "Parser" [
--    testCompressSpaces ]

