module UnitTest (
    -- Data.
    TestFailure(..), TestResult(..), TestCase,
    -- Functions.
    runTest, testSuite, testCase,
    check, checkEq,
    checkJust, checkNothing,
    checkLeft, checkRight,
    checkNull, checkNotNull, checkLength,
    -- Testing.
    testModule
) where

import List

------------------------------------------------------------------------
-- TestFailure definitions.
------------------------------------------------------------------------

data TestFailure = TestFailure {
    failureLocation :: [String],
    failureMessage  :: String
}

instance Show TestFailure where
    showsPrec _ (TestFailure location message) x 
        = printLocation location ++ ": " ++ message ++ x

printLocation :: [String] -> String
printLocation []       = "<unknown>"
printLocation location = concat (intersperse "." (reverse location))

------------------------------------------------------------------------
-- TestResult definitions.
------------------------------------------------------------------------

data TestResult = TestResult { 
    -- Current test location (based on testCase/testSuite names).
    testLocation  :: [String],

    -- Number of test cases executed.
    testCaseCount :: Integer,

    -- Failure information for every failed test case.
    testFailures  :: [TestFailure] 
}

instance Show TestResult where
    showsPrec _ (TestResult _ count []) x
        = "OK (" ++ show count ++ " test cases)\n" ++ x
    showsPrec _ (TestResult _ count failures) x
        = "FAILURES (" ++ show count ++ " test cases; " 
          ++ show (length failures) ++ " failures)\n\n"
          ++ unlines (map show (reverse failures)) ++ x

------------------------------------------------------------------------
-- Monad definitions.
------------------------------------------------------------------------

newtype TestCase a  = TestCase (TestResult -> (TestResult, Maybe a))

instance Monad TestCase where

    -- Run TEST and apply the result to K if TEST is successful.
    test >>= k = TestCase (\state ->
                      let (state', value) = run test state
                      in  case value of 
                              Just a  -> run (k a) state'
                              Nothing -> (state', Nothing))

    return value = TestCase (\state -> (state, Just value))

    fail message = TestCase (\state -> 
        (state { testFailures = (TestFailure (testLocation state) message 
                                 : testFailures state) },
         Nothing))

-- Run the TEST with STATE.
run :: TestCase a -> TestResult -> (TestResult, Maybe a)
run (TestCase test) state = test state

-- Update the test result using FUNCTION.
modifyResult :: (TestResult -> TestResult) -> TestCase ()
modifyResult function = TestCase (\state -> (function state, Just ()))

-- Query the test result using FUNCTION
queryResult :: (TestResult -> a) -> TestCase a
queryResult function = TestCase (\state -> (state, Just (function state)))

-- Increment the test case count.
incrementCount :: TestCase ()
incrementCount = modifyResult (\state -> 
    state { testCaseCount = 1 + testCaseCount state })

-- Add NAME to the current location.
insertName :: String -> TestCase ()
insertName name = modifyResult (\state -> 
    state { testLocation = name : testLocation state })

-- Remove the last name from the current location.
removeName :: TestCase ()
removeName = modifyResult (\state -> 
    state { testLocation = tail (testLocation state) })

-- Retrieve the current name.
getName :: TestCase String
getName = queryResult (\state -> head (testLocation state))

-- Run TEST and return result.  On failure run HANDLER and return
-- result.
catchFailure :: TestCase a -> TestCase a -> TestCase a
catchFailure test handler
    = TestCase (\state -> let (state', value) = run test state
                          in  case value of
                                  Just a  -> (state', Just a)
                                  Nothing -> run handler state')

-- Run TEST and ignore failures.
ignoreFailure :: TestCase () -> TestCase ()
ignoreFailure test = catchFailure test (return ())

initialState :: TestResult
initialState = TestResult { 
    testLocation  = [], 
    testCaseCount = 0, 
    testFailures  = [] 
}

-- Run TEST and retrieve the test results.
runTest :: TestCase a -> TestResult
runTest test = fst (run test initialState)

-- Run all TESTS using location NAME and ignore any failures.
testSuite :: String -> [TestCase ()] -> TestCase ()
testSuite name tests
    = do insertName name
         sequence_ (map ignoreFailure tests)
         removeName

-- Run TEST using location NAME and ignore failures.
testCase :: String -> TestCase () -> TestCase ()
testCase name test
    = do incrementCount
         insertName name
         ignoreFailure test
         removeName


------------------------------------------------------------------------
-- Checkion functions.
------------------------------------------------------------------------

check :: String -> Bool -> TestCase ()
check message False = fail message
check message True  = return ()

checkEq :: (Eq a, Show a) => String -> a -> a -> TestCase ()
checkEq message expected actual
    | expected == actual = return ()
    | otherwise = fail (message
                        ++ " expected: " ++ show expected
                        ++ " actual: "   ++ show actual)

checkJust :: String -> Maybe a -> TestCase ()
checkJust message (Just _) = return ()
checkJust message Nothing  = fail message

checkNothing :: Show a => String -> Maybe a -> TestCase ()
checkNothing message v@(Just _) = fail (message ++ ": " ++ show v)
checkNothing message Nothing    = return ()

checkLeft :: (Show a, Show b) => String -> Either a b -> TestCase ()
checkLeft message (Left _)    = return ()
checkLeft message v@(Right _) = fail (message ++ ": " ++ show v)

checkRight :: (Show a, Show b) => String -> Either a b -> TestCase ()
checkRight message v@(Left _) = fail (message ++ ": " ++ show v)
checkRight message (Right _)  = return ()

checkNull :: Show a => String -> [a] -> TestCase ()
checkNull message []   = return ()
checkNull message list = fail (message ++ ": " ++ show list)

checkNotNull :: String -> [a] -> TestCase ()
checkNotNull message []   = fail message
checkNotNull message list = return ()

checkLength :: String -> Int -> [a] -> TestCase ()
checkLength message expected list = checkEq message expected (length list)

------------------------------------------------------------------------
-- Tests
--
-- Run using 'runTest testModule'
------------------------------------------------------------------------

testModule :: TestCase ()
testModule = testSuite "UnitTest" [ 
    testReturn,
    testFail,
    testIncrementCount, 
    testCheck,
    testGetName ]

testReturn :: TestCase ()
testReturn = testCase "return" (
    do let result = snd $ run (return ()) initialState
       checkEq "result incorrect" (Just ()) result
       let result = snd $ run (return "value") initialState
       checkEq "result incorrect" (Just "value") result)

testFail :: TestCase ()
testFail = testCase "fail" (
    do let (state, result) = run (fail "error") initialState
       checkLength "no failure reported" 1 (testFailures state)
       checkNothing "no result failure" (result :: Maybe String)
       let state' = fst $ run (fail "another") state
       checkLength "failure not added" 2 (testFailures state'))

testIncrementCount :: TestCase ()
testIncrementCount = testCase "incrementCount" (
    do checkEq "initial count incorrect" 0 (testCaseCount initialState)
       let state = fst $ run incrementCount initialState
       checkEq "count not incremented" 1 (testCaseCount state))

testCheck :: TestCase ()
testCheck = testCase "check" (
    do let state = fst $ run (check "message" True) initialState
       checkNull "check True failed" (testFailures state)
       let state = fst $ run (check "message" False) initialState
       checkLength "check False succeeded" 1 (testFailures state))

testGetName :: TestCase ()
testGetName = testCase "getName" (
    do name <- getName
       checkEq "name incorrect" "getName" name)
